import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions
import time
from lal import GreenwichMeanSiderealTime
from astropy.time import Time
from astropy import coordinates as coord
import corner
import os
import shutil
import h5py
import json
import sys
from universal_divergence import estimate
import pandas as pd
import bilby
from bilby.core.prior import Uniform

def get_param_index(all_pars,pars,sky_extra=None):
    """ 
    Get the list index of requested source parameter types
    """
    # identify the indices of wrapped and non-wrapped parameters - clunky code
    mask = []
    idx = []

    # loop over inference params
    for i,p in enumerate(all_pars):

        # loop over wrapped params
        flag = False
        for q in pars:
            if p==q:
                flag = True    # if inf params is a wrapped param set flag

        # record the true/false value for this inference param
        if flag==True:
            mask.append(True)
            idx.append(i)
        elif flag==False:
            mask.append(False)

    if sky_extra is not None:
        if sky_extra:
            mask.append(True)
            idx.append(len(all_pars))
        else:
            mask.append(False)

    return mask, idx, np.sum(mask)

def convert_ra_to_hour_angle(data, params, pars, single=False):
    """
    Converts right ascension to hour angle and back again
    """

    greenwich = coord.EarthLocation.of_site('greenwich')
    t = Time(params['ref_geocent_time'], format='gps', location=greenwich)
    t = t.sidereal_time('mean', 'greenwich').radian

    # compute single instance
    if single:
        return t - data

    for i,k in enumerate(pars):
        if k == 'ra':
            ra_idx = i

    # Check if RA exist
    try:
        ra_idx
    except NameError:
        print('...... RA is fixed. Not converting RA to hour angle.')
    else:
        # Iterate over all training samples and convert to hour angle
        for i in range(data.shape[0]):
            data[i,ra_idx] = t - data[i,ra_idx]

    return data

def convert_hour_angle_to_ra(data, params, pars, single=False):
    """
    Converts right ascension to hour angle and back again
    """
    greenwich = coord.EarthLocation.of_site('greenwich')
    t = Time(params['ref_geocent_time'], format='gps', location=greenwich)
    t = t.sidereal_time('mean', 'greenwich').radian

    # compute single instance
    if single:
        return np.remainder(t - data,2.0*np.pi)

    for i,k in enumerate(pars):
        if k == 'ra':
            ra_idx = i

    # Check if RA exist
    try:
        ra_idx
    except NameError:
        print('...... RA is fixed. Not converting RA to hour angle.')
    else:
        # Iterate over all training samples and convert to hour angle
        for i in range(data.shape[0]):
            data[i,ra_idx] = np.remainder(t - data[i,ra_idx],2.0*np.pi)

    return data

def load_data(params,bounds,fixed_vals,input_dir,inf_pars,test_data=False,silent=False):
    """ Function to load either training or testing data.
    """

    # Get list of all training/testing files and define dictionary to store values in files
    if type("%s" % input_dir) is str:
        dataLocations = ["%s" % input_dir]
    else:
        print('ERROR: input directory not a string')
        exit(0)

    # Sort files from first generated to last generated
    filenames = sorted(os.listdir(dataLocations[0]))

    # If loading by chunks, randomly shuffle list of training/testing filenames
    if params['load_by_chunks'] == True and not test_data:
        nfiles = np.min([int(params['load_chunk_size']/float(params['tset_split'])),len(filenames)])
        files_idx = np.random.randint(0,len(filenames),nfiles) 
        filenames= np.array(filenames)[files_idx]
        if not silent:
            print('...... shuffled filenames since we are loading in by chunks')

    # Iterate over all training/testing files and store source parameters, time series and SNR info in dictionary
    data={'x_data': [], 'y_data_noisefree': [], 'y_data_noisy': [], 'rand_pars': [], 'snrs': []}
    for filename in filenames:
        try:
            data['x_data'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['x_data'][:])
            data['y_data_noisefree'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisefree'][:])
            if test_data:
                data['y_data_noisy'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['y_data_noisy'][:])
            data['rand_pars'] = h5py.File(dataLocations[0]+'/'+filename, 'r')['rand_pars'][:]
            data['snrs'].append(h5py.File(dataLocations[0]+'/'+filename, 'r')['snrs'][:])
            if not silent:
                print('...... Loaded file ' + dataLocations[0] + '/' + filename)
        except OSError:
            print('Could not load requested file')
            continue
    ndet = np.array(data['y_data_noisefree']).ndim
    if ndet == 3:
        data['y_data_noisefree'] = np.expand_dims(np.array(data['y_data_noisefree']),axis=0)
        data['y_data_noisy'] = np.expand_dims(np.array(data['y_data_noisy']),axis=0)
    data['x_data'] = np.concatenate(np.array(data['x_data']), axis=0).squeeze()
    data['y_data_noisefree'] = np.transpose(np.concatenate(np.array(data['y_data_noisefree']), axis=0),[0,2,1])
    if test_data:
        data['y_data_noisy'] = np.transpose(np.concatenate(np.array(data['y_data_noisy']), axis=0),[0,2,1])
    data['snrs'] = np.concatenate(np.array(data['snrs']), axis=0)

    # convert ra to hour angle
    data['x_data'] = convert_ra_to_hour_angle(data['x_data'], params, params['rand_pars'])

    # Normalise the source parameters
    for i,k in enumerate(data['rand_pars']):
        par_min = k.decode('utf-8') + '_min'
        par_max = k.decode('utf-8') + '_max'
        # normalize by bounds
        data['x_data'][:,i]=(data['x_data'][:,i] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])

    # extract inference parameters from all source parameters loaded earlier
    idx = []
    infparlist = ''
    for k in inf_pars:
        infparlist = infparlist + k + ', '
        for i,q in enumerate(data['rand_pars']):
            m = q.decode('utf-8')
            if k==m:
                idx.append(i)
    data['x_data'] = tf.cast(data['x_data'][:,idx],dtype=tf.float32)
    if not silent:
        print('...... {} will be inferred'.format(infparlist))

    return data['x_data'], tf.cast(data['y_data_noisefree'],dtype=tf.float32), tf.cast(data['y_data_noisy'],dtype=tf.float32), data['snrs'].reshape(-1,ndet)

def load_samples(params,sampler):
    """
    read in pre-computed posterior samples
    """
    if type("%s" % params['pe_dir']) is str:
        # load generated samples back in
        dataLocations = '%s_%s' % (params['pe_dir'],sampler)
        print('... looking in {} for posterior samples'.format(dataLocations))
    else:
        print('ERROR: input samples directory not a string')
        exit(0)

    # Iterate over requested number of testing samples to use
    for i in range(params['r']):

        filename = '%s/%s_%d.h5py' % (dataLocations,params['bilby_results_label'],i)
        if not os.path.isfile(filename):
            print('... unable to find file {}. Exiting.'.format(filename))
            exit(0)

        print('... Loading test sample -> ' + filename)
        data_temp = {}
        n = 0

        # Retrieve all source parameters to do inference on
        for q in params['bilby_pars']:
            p = q + '_post'
            par_min = q + '_min'
            par_max = q + '_max'
            data_temp[p] = h5py.File(filename, 'r')[p][:]
            if p == 'geocent_time_post':
                data_temp[p] = data_temp[p] - params['ref_geocent_time']
            data_temp[p] = (data_temp[p] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])
            Nsamp = data_temp[p].shape[0]
            n = n + 1
        print('... read in {} samples from {}'.format(Nsamp,filename))

        # place retrieved source parameters in numpy array rather than dictionary
        j = 0
        XS = np.zeros((Nsamp,n))
        for p,d in data_temp.items():
            XS[:,j] = d
            j += 1
        print('... put the samples in an array')

        # Append test sample posteriors to existing array of other test sample posteriors
        rand_idx_posterior = np.random.randint(0,Nsamp,params['n_samples'])
        if i == 0:
            XS_all = np.expand_dims(XS[rand_idx_posterior,:], axis=0)
        else:
            XS_all = np.vstack((XS_all,np.expand_dims(XS[rand_idx_posterior,:], axis=0)))
        print('... appended {} samples to the total'.format(params['n_samples']))
    return XS_all

def plot_losses(train_loss, val_loss, epoch, run='testing'):
    """
    plots the losses
    """
    plt.figure()
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,0],'b',label='RECON')
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,1],'r',label='KL')
    plt.semilogx(np.arange(1,epoch+1),train_loss[:epoch,2],'g',label='TOTAL')
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,0],'--b',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,1],'--r',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_loss[:epoch,2],'--g',alpha=0.5)
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.legend()
    plt.grid()
    plt.ylim([np.min(1.1*train_loss[int(0.1*epoch):epoch,:]),np.max(1.1*train_loss[int(0.1*epoch):epoch,:])])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/loss.png' % (run))
    plt.close()

    # save loss data to text file
    loss_file = '/data/www.astro/chrism/vitamin_c/{}/loss.txt'.format(run)
    f = open(loss_file,'a+')
    data = tf.reshape(np.concatenate([train_loss[epoch-1,:],val_loss[epoch-1,:]],axis=0),(1,6))
    f.write('%f %f %f %f %f %f\n' % (data[0,0],data[0,1],data[0,2],data[0,3],data[0,4],data[0,5]))
    #np.savetxt(loss_file,data)
    f.close()

 
def plot_KL(KL_samples, step, run='testing'):
    """
    plots the KL evolution
    """
    # arrives in shape n_kl,n_test,3
    N = KL_samples.shape[0]
    KL_samples = np.transpose(KL_samples,[2,1,0])   # re-order axes
    ls = ['-','--',':']
    c = ['C0','C1','C2','C3']
    fig, axs = plt.subplots(3, sharex=True, sharey=True, figsize=(6.4,14.4))
    for i,kl_s in enumerate(KL_samples):   # loop over samplers
        for j,kl in enumerate(kl_s):     # loop over test cases
            axs[i].semilogx(np.arange(1,N+1)*step,kl,ls[i],color=c[j])
            axs[i].plot(N*step,kl[-1],'.',color=c[j])
            axs[i].grid()
    plt.xlabel('epoch')
    plt.ylabel('KL')
    plt.ylim([-0.2,1.0])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/kl.png' % (run))
    plt.close()


def plot_posterior(samples,x_truth,snrs,epoch,idx,run='testing',all_other_samples=None):
    """
    plots the posteriors
    """

    # trim samples from outside the cube
    mask = []
    for s in samples:
        if (np.all(s>=0.0) and np.all(s<=1.0)):
            mask.append(True)
        else:
            mask.append(False)
    samples = tf.boolean_mask(samples,mask,axis=0)
    print('identified {} good samples'.format(samples.shape[0]))
    print(np.array(all_other_samples).shape)
    if samples.shape[0]<100:
        return [-1.0,-1.0,-1.0]

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')
    hist_kwargs_other2 = dict(density=True,color='tab:green')

    if all_other_samples is not None:
        KL_est = []
        for i, other_samples in enumerate(all_other_samples):
            true_post = np.zeros([other_samples.shape[0],bilby_ol_len])
            true_x = np.zeros(inf_ol_len)
            true_XS = np.zeros([samples.shape[0],inf_ol_len])
            ol_pars = []
            cnt = 0
            for inf_idx,bilby_idx in zip(inf_ol_idx,bilby_ol_idx):
                inf_par = params['inf_pars'][inf_idx]
                bilby_par = params['bilby_pars'][bilby_idx]
                true_XS[:,cnt] = (samples[:,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
                true_post[:,cnt] = (other_samples[:,bilby_idx] * (bounds[bilby_par+'_max'] - bounds[bilby_par+'_min'])) + bounds[bilby_par + '_min']
                true_x[cnt] = (x_truth[inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
                ol_pars.append(inf_par)
                cnt += 1
            parnames = []
            for k_idx,k in enumerate(params['rand_pars']):
                if np.isin(k, ol_pars):
                    parnames.append(params['corner_labels'][k])

            # convert to RA
            true_XS = convert_hour_angle_to_ra(true_XS,params,ol_pars)
            true_x = convert_hour_angle_to_ra(np.reshape(true_x,[1,true_XS.shape[1]]),params,ol_pars).flatten()
            old_true_post = true_post                 

            samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_epoch_{}_event_{}_vit.txt'.format(run,epoch,idx)
            np.savetxt(samples_file,true_XS)

            # compute KL estimate
            idx1 = np.random.randint(0,true_XS.shape[0],2000)
            idx2 = np.random.randint(0,true_post.shape[0],2000)
            try:
                current_KL = 0.5*(estimate(true_XS[idx1,:],true_post[idx2,:],n_jobs=4) + estimate(true_post[idx2,:],true_XS[idx1,:],n_jobs=4))
            except:
                current_KL = -1.0
                pass
            KL_est.append(current_KL)

            other_samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_epoch_{}_event_{}_{}.txt'.format(run,epoch,idx,i)
            np.savetxt(other_samples_file,true_post)

            if i==0:
                figure = corner.corner(true_post, **defaults_kwargs,labels=parnames,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
            else:

                # compute KL estimate
                idx1 = np.random.randint(0,old_true_post.shape[0],2000)
                idx2 = np.random.randint(0,true_post.shape[0],2000)
                try:
                    current_KL = 0.5*(estimate(old_true_post[idx1,:],true_post[idx2,:],n_jobs=4) + estimate(true_post[idx2,:],old_true_post[idx1,:],n_jobs=4))
                except:
                    current_KL = -1.0
                    pass
                KL_est.append(current_KL)

                corner.corner(true_post,**defaults_kwargs,
                           color='tab:green',
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs_other2)
        
        for j,KL in enumerate(KL_est):    
            plt.annotate('KL = {:.3f}'.format(KL),(0.2,0.95-j*0.02),xycoords='figure fraction',fontsize=18)

        snr_tot = np.sqrt(np.sum(snrs**2))
        plt.annotate('SNR = {:.3f} [{:.3f},{:.3f},{:.3f}]'.format(snr_tot,snrs[0],snrs[1],snrs[2]),(0.6,0.95),xycoords='figure fraction',fontsize=18)
        corner.corner(true_XS,**defaults_kwargs,
                           color='tab:red',
                           fill_contours=True, truths=true_x,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/comp_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()
        return KL_est

    else:
        # Get corner parnames to use in plotting labels
        parnames = []
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, params['inf_pars']):
                parnames.append(params['corner_labels'][k])
        # un-normalise full inference parameters
        full_true_x = np.zeros(len(params['inf_pars']))
        new_samples = np.zeros([samples.shape[0],len(params['inf_pars'])])
        for inf_par_idx,inf_par in enumerate(params['inf_pars']):
            new_samples[:,inf_par_idx] = (samples[:,inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
            full_true_x[inf_par_idx] = (x_truth[inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
        new_samples = convert_hour_angle_to_ra(new_samples,params,params['inf_pars'])
        full_true_x = convert_hour_angle_to_ra(np.reshape(full_true_x,[1,samples.shape[1]]),params,params['inf_pars']).flatten()       

        figure = corner.corner(new_samples,**defaults_kwargs,labels=parnames,
                           color='tab:red',
                           fill_contours=True, truths=full_true_x,
                           show_titles=True, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/full_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()
    return -1.0

def plot_latent(mu_r1, z_r1, mu_q, z_q, epoch, idx, run='testing'):

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')

    figure = corner.corner(z_q, **defaults_kwargs,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
    corner.corner(z_r1,**defaults_kwargs,
                           color='tab:red',
                           fill_contours=True,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
    # Extract the axes
    z_dim = z_r1.shape[1]
    axes = np.array(figure.axes).reshape((z_dim, z_dim))

    # Loop over the histograms
    for yi in range(z_dim):
        for xi in range(yi):
            ax = axes[yi, xi]
            ax.plot(mu_r1[0,:,xi], mu_r1[0,:,yi], "sr")
            ax.plot(mu_q[0,xi], mu_q[0,yi], "sb")
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

params = '/home/chrism/vitamin_c/params.json'
bounds = '/home/chrism/vitamin_c//bounds.json'
fixed_vals = '/home/chrism/vitamin_c/fixed_vals.json'
#run = time.ctime().replace(' ', '-')
run = time.strftime('%y-%m-%d-%X-%Z')
EPS = 1e-3

path = os.path.join('/data/www.astro/chrism/vitamin_c', run) 
os.mkdir(path) 
shutil.copy('./vitamin_c.py',path)
shutil.copy('./params.json',path)

# Load parameters files
with open(params, 'r') as fp:
    params = json.load(fp)
with open(bounds, 'r') as fp:
    bounds = json.load(fp)
with open(fixed_vals, 'r') as fp:
    fixed_vals = json.load(fp)

# if doing hour angle, use hour angle bounds on RA
bounds['ra_min'] = convert_ra_to_hour_angle(bounds['ra_min'],params,None,single=True)
bounds['ra_max'] = convert_ra_to_hour_angle(bounds['ra_max'],params,None,single=True)
print('... converted RA bounds to hour angle')
inf_ol_mask, inf_ol_idx, inf_ol_len = get_param_index(params['inf_pars'],params['bilby_pars'])
bilby_ol_mask, bilby_ol_idx, bilby_ol_len = get_param_index(params['bilby_pars'],params['inf_pars'])

# identify the indices of different sets of physical parameters
vonmise_mask, vonmise_idx_mask, vonmise_len = get_param_index(params['inf_pars'],params['vonmise_pars'])
gauss_mask, gauss_idx_mask, gauss_len = get_param_index(params['inf_pars'],params['gauss_pars'])
sky_mask, sky_idx_mask, sky_len = get_param_index(params['inf_pars'],params['sky_pars'])
ra_mask, ra_idx_mask, ra_len = get_param_index(params['inf_pars'],['ra'])
dec_mask, dec_idx_mask, dec_len = get_param_index(params['inf_pars'],['dec'])
m1_mask, m1_idx_mask, m1_len = get_param_index(params['inf_pars'],['mass_1'])
m2_mask, m2_idx_mask, m2_len = get_param_index(params['inf_pars'],['mass_2'])
#idx_mask = np.argsort(gauss_idx_mask + vonmise_idx_mask + m1_idx_mask + m2_idx_mask + sky_idx_mask) # + dist_idx_mask)
idx_mask = np.argsort(m1_idx_mask + m2_idx_mask + gauss_idx_mask + vonmise_idx_mask) # + sky_idx_mask)
dist_mask, dist_idx_mask, dist_len = get_param_index(params['inf_pars'],['luminosity_distance'])
not_dist_mask, not_dist_idx_mask, not_dist_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
phase_mask, phase_idx_mask, phase_len = get_param_index(params['inf_pars'],['phase'])
not_phase_mask, not_phase_idx_mask, not_phase_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
geocent_mask, geocent_idx_mask, geocent_len = get_param_index(params['inf_pars'],['geocent_time'])
not_geocent_mask, not_geocent_idx_mask, not_geocent_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','phase','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
xyz_mask, xyz_idx_mask, xyz_len = get_param_index(params['inf_pars'],['luminosity_distance','ra','dec'])
not_xyz_mask, not_xyz_idx_mask, not_xyz_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
idx_xyz_mask = np.argsort(xyz_idx_mask + not_xyz_idx_mask)
idx_dist_mask = np.argsort(not_dist_idx_mask + dist_idx_mask)
idx_phase_mask = np.argsort(not_phase_idx_mask + phase_idx_mask)
idx_geocent_mask = np.argsort(not_geocent_idx_mask + geocent_idx_mask)
print(xyz_mask)
print(not_xyz_mask)
print(idx_xyz_mask)
#masses_len = m1_len + m2_len
print(params['inf_pars'])
print(vonmise_mask,vonmise_idx_mask)
print(gauss_mask,gauss_idx_mask)
print(m1_mask,m1_idx_mask)
print(m2_mask,m2_idx_mask)
print(sky_mask,sky_idx_mask)
print(idx_mask)

# define which gpu to use during training
gpu_num = str(params['gpu_num'])   
os.environ["CUDA_VISIBLE_DEVICES"]=gpu_num
print('... running on GPU {}'.format(gpu_num))

# Let GPU consumption grow as needed
config = tf.compat.v1.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.compat.v1.Session(config=config)
print('... letting GPU consumption grow as needed')

# load the data
x_data_train, y_data_train, _, snrs_train = load_data(params,bounds,fixed_vals,params['train_set_dir'],params['inf_pars'])

#print(x_data_train[1,:])
#plt.figure()
#plt.plot(y_data_train[1,:,0],'b')

# augment the data for this chunk - randomize phase and arrival time
#old_phase = bounds['phase_min'] + tf.boolean_mask(x_data_train,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
#new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
#new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
#x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_phase_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
#old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x_data_train,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
#new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
#new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
#x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_geocent_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
#phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
#phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y_data_train)[1]/2 + 1))
#fvec = tf.range(params['ndata']/2 + 1)/params['duration']
#time_correction = -2.0*np.pi*fvec*(new_geocent-old_geocent)
#time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
#time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
#y_data_train_fft = tf.signal.rfft(tf.transpose(y_data_train,[0,2,1]))*phase_correction*time_correction
#y_data_train = tf.transpose(tf.signal.irfft(y_data_train_fft),[0,2,1])

#print(x_data_train[1,:])
#print(old_geocent[1],new_geocent[1])
#plt.plot(y_data_train[1,:,0],'r')
#plt.savefig('/data/www.astro/chrism/vitamin_c/time_phase_shift.png')
#plt.close()
#exit(0)

x_data_val, y_data_val, _, snrs_val = load_data(params,bounds,fixed_vals,params['val_set_dir'],params['inf_pars'])
x_data_test, y_data_test_noisefree, y_data_test, snrs_test = load_data(params,bounds,fixed_vals,params['test_set_dir'],params['inf_pars'],test_data=True)
x_data_pp, y_data_pp_noisefree, y_data_pp, snrs_pp = load_data(params,bounds,fixed_vals,params['pp_set_dir'],params['inf_pars'],test_data=True)

print(x_data_val.shape)

snrs_test = np.reshape(snrs_test,[-1,3])

plt.figure()
plt.hist(np.sqrt(np.sum(snrs_train**2,axis=1)),100,density=True,histtype='stepfilled',alpha=0.5)
plt.hist(np.sqrt(np.sum(snrs_val**2,axis=1)),100,density=True,histtype='stepfilled',alpha=0.5)
plt.plot(np.sqrt(np.sum(snrs_pp**2,axis=1)),np.zeros(snrs_pp.shape[0]),'ok')
plt.plot(np.sqrt(np.sum(snrs_test**2,axis=1)),np.zeros(snrs_test.shape[0]),'sr')
plt.plot(np.sqrt(np.sum(snrs_pp[:4,:]**2,axis=1)),np.zeros(4),'ob')
plt.savefig('/data/www.astro/chrism/vitamin_c/%s/snrs.png' % (run))
plt.close()

# load precomputed samples
#bilby_samples = np.array([load_samples(params,'dynesty'),load_samples(params,'ptemcee')])
bilby_samples = np.array([load_samples(params,'dynesty')])
print(bilby_samples.shape)

train_size = params['load_chunk_size']
batch_size = 256
val_size = batch_size*19
test_size = 4
plot_cadence = 500

train_dataset = (tf.data.Dataset.from_tensor_slices((x_data_train,y_data_train))
                 .shuffle(train_size).batch(batch_size))
val_dataset = (tf.data.Dataset.from_tensor_slices((x_data_val,y_data_val))
                .take(val_size).batch(batch_size))
print(x_data_test.shape,y_data_test.shape,snrs_test.shape)
test_dataset = (tf.data.Dataset.from_tensor_slices((x_data_test,y_data_test,snrs_test))
                .batch(1))
print(x_data_pp.shape,y_data_pp.shape,snrs_pp.shape)
pp_dataset = (tf.data.Dataset.from_tensor_slices((x_data_pp,y_data_pp,snrs_pp))
                .batch(1))

class CVAE(tf.keras.Model):
    """Convolutional variational autoencoder."""

    def __init__(self, x_dim, y_dim, n_channels, z_dim, n_modes):
        super(CVAE, self).__init__()
        self.z_dim = z_dim
        self.n_modes = n_modes
        self.x_modes = 1   # hardcoded for testing
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.n_channels = n_channels
        self.act = 'relu'

        #r1_input_y = tf.keras.Input(shape=(self.y_dim, self.n_channels))
 
        # data reduction
        #layer_0 = tf.keras.layers.Conv1D(filters=64, kernel_size=5, strides=1, padding='same', activation='relu')(r1_input_y)
        #layer_1 = tf.keras.layers.Conv1D(filters=64, kernel_size=5, strides=2, padding='same')(layer_0)
        #layer_2 = tf.keras.layers.Conv1D(filters=64, kernel_size=5, strides=2, padding='same', use_bias=False)(r1_input_y)
        #layer_12 = tf.keras.activations.relu(layer_1 + layer_2)
        #layer_3 = tf.keras.layers.Conv1D(filters=1, kernel_size=5, strides=2, padding='same')(layer_12)
        #layer_4 = tf.keras.layers.Conv1D(filters=1, kernel_size=5, strides=4, padding='same', use_bias=False)(layer_0)
        #layer_5 = tf.keras.layers.Conv1D(filters=1, kernel_size=5, strides=4, padding='same', use_bias=False)(r1_input_y)
        #layer_red = tf.keras.activations.relu(layer_3 + layer_4 + layer_5)
       

        #layer_0 = tf.keras.layers.Conv1D(filters=64, kernel_size=7, strides=2, padding='same', activation='relu')(r1_input_y)
        #layer_0 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(layer_0)
        #layer_0 = tf.nn.local_response_normalization(layer_0, depth_radius=5, bias=1, alpha=1, beta=0.5, name=None)
        #layer_0 = tf.keras.layers.Conv1D(filters=192, kernel_size=1, strides=1, padding='same', activation='relu')(layer_0)
        #layer_0 = tf.keras.layers.Conv1D(filters=192, kernel_size=3, strides=1, padding='same', activation='relu')(layer_0)
        #layer_0 = tf.nn.local_response_normalization(layer_0, depth_radius=5, bias=1, alpha=1, beta=0.5, name=None)
        #layer_0 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(layer_0)

        ### 1st inception
        #inc_1a = tf.keras.layers.Conv1D(filters=96, kernel_size=1, padding='same', activation='relu')(layer_0)
        #inc_1a = tf.keras.layers.Conv1D(filters=128, kernel_size=3, padding='same', activation='relu')(inc_1a)
        #inc_1b = tf.keras.layers.Conv1D(filters=16, kernel_size=1, padding='same', activation='relu')(layer_0)
        #inc_1b = tf.keras.layers.Conv1D(filters=32, kernel_size=5, padding='same', activation='relu')(inc_1b)
        #inc_1c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(layer_0)
        #inc_1c = tf.keras.layers.Conv1D(filters=32, kernel_size=1, padding='same', activation='relu')(inc_1c)
        #inc_1d = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(layer_0)
        #inc_1_out = tf.keras.layers.concatenate([inc_1a, inc_1b, inc_1c, inc_1d], axis=2)
        #inc_1_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_1_out)

        # 2nd inception
        #inc_2a = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_1_out)
        #inc_2a = tf.keras.layers.Conv1D(filters=192, kernel_size=3, padding='same', activation='relu')(inc_2a)
        #inc_2b = tf.keras.layers.Conv1D(filters=32, kernel_size=1, padding='same', activation='relu')(inc_1_out)
        #inc_2b = tf.keras.layers.Conv1D(filters=96, kernel_size=5, padding='same', activation='relu')(inc_2b)
        #inc_2c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_1_out)
        #inc_2c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(inc_2c)
        #inc_2d = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_1_out)
        #inc_2_out = tf.keras.layers.concatenate([inc_2a, inc_2b, inc_2c, inc_2d], axis=2)
        #inc_2_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_2_out)

        # 3rd inception
        #inc_3a = tf.keras.layers.Conv1D(filters=96, kernel_size=1, padding='same', activation='relu')(inc_2_out)
        #inc_3a = tf.keras.layers.Conv1D(filters=208, kernel_size=3, padding='same', activation='relu')(inc_3a)
        #inc_3b = tf.keras.layers.Conv1D(filters=16, kernel_size=1, padding='same', activation='relu')(inc_2_out)
        #inc_3b = tf.keras.layers.Conv1D(filters=48, kernel_size=5, padding='same', activation='relu')(inc_3b)
        #inc_3c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_2_out)
        #inc_3c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(inc_3c)
        #inc_3d = tf.keras.layers.Conv1D(filters=192, kernel_size=1, padding='same', activation='relu')(inc_2_out)
        #inc_3_out = tf.keras.layers.concatenate([inc_3a, inc_3b, inc_3c, inc_3d], axis=2)
        ##inc_3_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_3_out)

        # 4th inception
        #inc_4a = tf.keras.layers.Conv1D(filters=112, kernel_size=1, padding='same', activation='relu')(inc_3_out)
        #inc_4a = tf.keras.layers.Conv1D(filters=224, kernel_size=3, padding='same', activation='relu')(inc_4a)
        #inc_4b = tf.keras.layers.Conv1D(filters=24, kernel_size=1, padding='same', activation='relu')(inc_3_out)
        #inc_4b = tf.keras.layers.Conv1D(filters=64, kernel_size=5, padding='same', activation='relu')(inc_4b)
        #inc_4c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_3_out)
        #inc_4c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(inc_4c)
        #inc_4d = tf.keras.layers.Conv1D(filters=160, kernel_size=1, padding='same', activation='relu')(inc_3_out)
        #inc_4_out = tf.keras.layers.concatenate([inc_4a, inc_4b, inc_4c, inc_4d], axis=2)
        #inc_4_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_4_out)

        # 5th inception
        #inc_5a = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_4_out)
        #inc_5a = tf.keras.layers.Conv1D(filters=256, kernel_size=3, padding='same', activation='relu')(inc_5a)
        #inc_5b = tf.keras.layers.Conv1D(filters=24, kernel_size=1, padding='same', activation='relu')(inc_4_out)
        #inc_5b = tf.keras.layers.Conv1D(filters=64, kernel_size=5, padding='same', activation='relu')(inc_5b)
        #inc_5c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_4_out)
        #inc_5c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(inc_5c)
        #inc_5d = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_4_out)
        #inc_5_out = tf.keras.layers.concatenate([inc_5a, inc_5b, inc_5c, inc_5d], axis=2)
        #inc_5_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_5_out)

        # 6th inception
        #inc_6a = tf.keras.layers.Conv1D(filters=144, kernel_size=1, padding='same', activation='relu')(inc_5_out)
        #inc_6a = tf.keras.layers.Conv1D(filters=288, kernel_size=3, padding='same', activation='relu')(inc_6a)
        #inc_6b = tf.keras.layers.Conv1D(filters=32, kernel_size=1, padding='same', activation='relu')(inc_5_out)
        #inc_6b = tf.keras.layers.Conv1D(filters=64, kernel_size=5, padding='same', activation='relu')(inc_6b)
        #inc_6c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_5_out)
        #inc_6c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, padding='same', activation='relu')(inc_6c)
        #inc_6d = tf.keras.layers.Conv1D(filters=112, kernel_size=1, padding='same', activation='relu')(inc_5_out)
        #inc_6_out = tf.keras.layers.concatenate([inc_6a, inc_6b, inc_6c, inc_6d], axis=2)
        #inc_6_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_6_out)

        # 7th inception
        #inc_7a = tf.keras.layers.Conv1D(filters=160, kernel_size=1, padding='same', activation='relu')(inc_6_out)
        #inc_7a = tf.keras.layers.Conv1D(filters=320, kernel_size=3, padding='same', activation='relu')(inc_7a)
        #inc_7b = tf.keras.layers.Conv1D(filters=32, kernel_size=1, padding='same', activation='relu')(inc_6_out)
        #inc_7b = tf.keras.layers.Conv1D(filters=128, kernel_size=5, padding='same', activation='relu')(inc_7b)
        #inc_7c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_6_out)
        #inc_7c = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_7c)
        #inc_7d = tf.keras.layers.Conv1D(filters=256, kernel_size=1, padding='same', activation='relu')(inc_6_out)
        #inc_7_out = tf.keras.layers.concatenate([inc_7a, inc_7b, inc_7c, inc_7d], axis=2)
        #inc_7_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_7_out)

        # 8th inception
        #inc_8a = tf.keras.layers.Conv1D(filters=160, kernel_size=1, padding='same', activation='relu')(inc_7_out)
        #inc_8a = tf.keras.layers.Conv1D(filters=320, kernel_size=3, padding='same', activation='relu')(inc_8a)
        #inc_8b = tf.keras.layers.Conv1D(filters=32, kernel_size=1, padding='same', activation='relu')(inc_7_out)
        #inc_8b = tf.keras.layers.Conv1D(filters=128, kernel_size=5, padding='same', activation='relu')(inc_8b)
        #inc_8c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_7_out)
        #inc_8c = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_8c)
        #inc_8d = tf.keras.layers.Conv1D(filters=256, kernel_size=1, padding='same', activation='relu')(inc_7_out)
        #inc_8_out = tf.keras.layers.concatenate([inc_8a, inc_8b, inc_8c, inc_8d], axis=2)
        #inc_8_out = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(inc_8_out)

        # 9th inception
        #inc_9a = tf.keras.layers.Conv1D(filters=192, kernel_size=1, padding='same', activation='relu')(inc_8_out)
        #inc_9a = tf.keras.layers.Conv1D(filters=384, kernel_size=3, padding='same', activation='relu')(inc_9a)
        #inc_9b = tf.keras.layers.Conv1D(filters=48, kernel_size=1, padding='same', activation='relu')(inc_8_out)
        #inc_9b = tf.keras.layers.Conv1D(filters=128, kernel_size=5, padding='same', activation='relu')(inc_9b)
        #inc_9c = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(inc_8_out)
        #inc_9c = tf.keras.layers.Conv1D(filters=128, kernel_size=1, padding='same', activation='relu')(inc_9c)
        #inc_9d = tf.keras.layers.Conv1D(filters=384, kernel_size=1, padding='same', activation='relu')(inc_8_out)
        #inc_9_out = tf.keras.layers.concatenate([inc_9a, inc_9b, inc_9c, inc_9d], axis=2)
        #inc_9_out = tf.keras.layers.AveragePooling1D(pool_size=8, strides=8, padding='same')(inc_9_out)
        #inc_9_out = tf.keras.layers.Dropout(0.4)(inc_9_out)

        #a = tf.keras.layers.Flatten()(mid_2)

              # the r1 encoder network
        #r1_input_y = tf.keras.Input(shape=(self.y_dim, self.n_channels))

        #a1 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(r1_input_y)
        #a1 = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='same', activation='relu')(a1)
        #a2 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(r1_input_y)
        #a2 = tf.keras.layers.Conv1D(filters=32, kernel_size=5, strides=1, padding='same', activation='relu')(a2)
        #a3 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(r1_input_y)
        #a3 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(a3)
        #a4 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(r1_input_y)
        #in1 = tf.keras.layers.concatenate([a1, a2, a3, a4], axis=2)
        #in1 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(in1)

        #a5 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in1)
        #a5 = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='same', activation='relu')(a5)
        #a6 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in1)
        #a6 = tf.keras.layers.Conv1D(filters=32, kernel_size=5, strides=1, padding='same', activation='relu')(a6)
        #a7 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(in1)
        #a7 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(a7)
        #a8 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in1)
        #in2 = tf.keras.layers.concatenate([a5, a6, a7, a8], axis=2)
        #in2 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(in2)

        #a9 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in2)
        #a9 = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='same', activation='relu')(a9)
        #a10 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in2)
        #a10 = tf.keras.layers.Conv1D(filters=32, kernel_size=5, strides=1, padding='same', activation='relu')(a10)
        #a11 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(in2)
        #a11 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(a11)
        #a12 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in2)
        #in3 = tf.keras.layers.concatenate([a9, a10, a11, a12], axis=2)
        ##in3 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(in3)

        #a13 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in3)
        #a13 = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=2, padding='same', activation='relu')(a13)
        #a14 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=1, padding='same', activation='relu')(in3)
        #a14 = tf.keras.layers.Conv1D(filters=32, kernel_size=5, strides=2, padding='same', activation='relu')(a14)
        #a15 = tf.keras.layers.MaxPooling1D(pool_size=3, strides=1, padding='same')(in3)
        #a15 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=2, padding='same', activation='relu')(a15)
        #a16 = tf.keras.layers.Conv1D(filters=32, kernel_size=1, strides=2, padding='same', activation='relu')(in3)
        #in4 = tf.keras.layers.concatenate([a13, a14, a15, a16], axis=2)
        #a = tf.keras.layers.MaxPooling1D(pool_size=3, strides=2, padding='same')(in4)

        # the r1 encoder network
        r1_input_y = tf.keras.Input(shape=(self.y_dim, self.n_channels))
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=64, strides=1, padding='same', activation=self.act)(r1_input_y)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(aa)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=32, strides=4, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,aa])
        #a = tf.keras.layers.Activation(self.act)(a)

        #a = tf.keras.layers.MaxPooling1D(pool_size=3,strides=3)(a)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=32, strides=1, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(ab)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=16, strides=2, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,ab])
        #a = tf.keras.layers.Activation(self.act)(a)

        #a = tf.keras.layers.MaxPooling1D(pool_size=2,strides=2)(a)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=16, strides=1, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(ac)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=16, strides=2, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,ac])
        #a = tf.keras.layers.Activation(self.act)(a)

        #a = tf.keras.layers.MaxPooling1D(pool_size=3,strides=3)(a)
        a = tf.keras.layers.Conv1D(filters=96, kernel_size=8, strides=2, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(ad)        
        #a = tf.keras.layers.Conv1D(filters=96, kernel_size=5, strides=1, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,ad])
        #a = tf.keras.layers.Activation(self.act)(a)

        #a = tf.keras.layers.MaxPooling1D(pool_size=2,strides=2)(a)
        #a = tf.keras.layers.Conv1D(filters=96, kernel_size=3, strides=1, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(ae)
        #a = tf.keras.layers.Conv1D(filters=96, kernel_size=3, strides=1, padding='same', activation=self.act)(a)
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,ae])
        #a = tf.keras.layers.Activation(self.act)(a)

        #a = tf.keras.layers.MaxPooling1D(pool_size=2,strides=2)(a)
        #af = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(a)
        #af = tf.keras.layers.BatchNormalization(trainable=False)(af)
        #a = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(af)      
        #a = tf.keras.layers.BatchNormalization(trainable=False)(a)
        #a = tf.keras.layers.Add()([a,af])
        #a = tf.keras.layers.Activation(self.act)(a)

        a = tf.keras.layers.Flatten()(a)
        a1 = tf.keras.layers.Dense(4096,activation=self.act)(a)
        a1 = tf.keras.layers.Dense(2048,activation=self.act)(a1)
        a1 = tf.keras.layers.Dense(1024,activation=self.act)(a1)
        a1 = tf.keras.layers.Dense(2*self.z_dim*self.n_modes + self.n_modes)(a1)
        self.encoder_r1 = tf.keras.Model(inputs=r1_input_y, outputs=a1)
        print(self.encoder_r1.summary())

        # the q encoder network
        q_input_x = tf.keras.Input(shape=(self.x_dim))
        c = tf.keras.layers.Flatten()(q_input_x)
        d = tf.keras.layers.concatenate([a,c])        
        e = tf.keras.layers.Dense(4096,activation=self.act)(d)
        e = tf.keras.layers.Dense(2048,activation=self.act)(e)
        e = tf.keras.layers.Dense(1024,activation=self.act)(e)
        e = tf.keras.layers.Dense(2*self.z_dim)(e)
        self.encoder_q = tf.keras.Model(inputs=[r1_input_y, q_input_x], outputs=e)
        print(self.encoder_q.summary())

        # the r2 decoder network
        r2_input_z = tf.keras.Input(shape=(self.z_dim))
        g = tf.keras.layers.Flatten()(r2_input_z)
        h = tf.keras.layers.concatenate([a,g])
        i = tf.keras.layers.Dense(4096,activation=self.act)(h)
        i = tf.keras.layers.Dense(2048,activation=self.act)(i)
        i = tf.keras.layers.Dense(1024,activation=self.act)(i)
        j = tf.keras.layers.Dense(2*self.x_dim)(i)
        self.decoder_r2 = tf.keras.Model(inputs=[r1_input_y, r2_input_z], outputs=j)
        print(self.decoder_r2.summary())

    def encode_r1(self, y=None):
        mean, logvar, weight = tf.split(self.encoder_r1(y), num_or_size_splits=[self.z_dim*self.n_modes, self.z_dim*self.n_modes,self.n_modes], axis=1)
        return tf.reshape(mean,[-1,self.n_modes,self.z_dim]), tf.reshape(logvar,[-1,self.n_modes,self.z_dim]), tf.reshape(weight,[-1,self.n_modes])
        #return tf.split(self.encoder_r1(y), num_or_size_splits=[self.z_dim*self.n_modes, self.z_dim*self.n_modes,self.n_modes], axis=1)

    def encode_q(self, x=None, y=None):
        #mean, logvar = tf.split(self.encoder_q([y,x]), num_or_size_splits=[self.z_dim,self.z_dim], axis=1)
        #return tf.reshape(mean,[-1,self.z_dim]), tf.reshape(logvar,[-1,self.z_dim])
        return tf.split(self.encoder_q([y,x]), num_or_size_splits=[self.z_dim,self.z_dim], axis=1)

    def decode_r2(self, y=None, z=None, apply_sigmoid=False):
        #return tf.split(self.decoder_r2([y,z]), num_or_size_splits=[self.x_dim,self.x_dim], axis=1)
        mean, logvar = tf.split(self.decoder_r2([y,z]), num_or_size_splits=[self.x_dim, self.x_dim], axis=1)
        return tf.reshape(mean,[-1,self.x_dim]), tf.reshape(logvar,[-1,self.x_dim])
        #return mean, logvar

optimizer = tf.keras.optimizers.Adam(1e-4)

def compute_loss(model, x, y, ramp=1.0, noiseamp=1.0):

    # randomise distance    
    old_d = bounds['luminosity_distance_min'] + tf.boolean_mask(x,dist_mask,axis=1)*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
    new_x = tf.random.uniform(shape=tf.shape(old_d), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
    new_d = bounds['luminosity_distance_min'] + new_x*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
    x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_dist_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_dist_mask),axis=1)
    dist_scale = tf.tile(tf.expand_dims(old_d/new_d,axis=1),(1,tf.shape(y)[1],1))

    # add noise and randomise distance again
    y = (y*dist_scale + noiseamp*tf.random.normal(shape=tf.shape(y), mean=0.0, stddev=1.0, dtype=tf.float32))/params['y_normscale']
    mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
    scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
    gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1,
            scale_diag=scale_r1))
    mean_q, logvar_q = model.encode_q(x=x,y=y)
    scale_q = EPS + tf.sqrt(tf.exp(logvar_q))
    mvn_q = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_q,
                          scale_diag=scale_q)
    #mvn_q = tfd.Normal(loc=mean_q,scale=scale_q)
    z_samp = mvn_q.sample()
    mean_r2, logvar_r2 = model.decode_r2(z=z_samp,y=y)
    scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))

    # SIMPLE RECON LOSS
    mvn_r2 = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_r2,
                          scale_diag=scale_r2)

    #gm_r2 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r2),
    #        components_distribution=tfd.MultivariateNormalDiag(
    #        loc=mean_r2,
    #        scale_diag=scale_r2))
    #gm_r2 = tfp.distributions.Beta(
    #        concentration1=mean_r2+EPS, 
    #        concentration0=logvar_r2+EPS, 
    #        validate_args=True, allow_nan_stats=False)
    # LOCALISATION
    #ra = 2*np.pi*tf.reshape(tf.boolean_mask(x,ra_mask,axis=1),[-1,1])       # convert the scaled 0->1 true RA value back to radians
    #dec = np.pi*(tf.reshape(tf.boolean_mask(x,dec_mask,axis=1),[-1,1]) - 0.5) # convert the scaled 0>1 true dec value back to radians
    #dist = tf.boolean_mask(x,dist_mask,axis=1)
    #xyz_actual = tf.reshape(tf.concat([dist*tf.cos(ra)*tf.cos(dec),dist*tf.sin(ra)*tf.cos(dec),dist*tf.sin(dec)],axis=1),[-1,3])   # construct the true parameter unit vector   
    #new_x = tf.gather(tf.concat([xyz_actual, tf.boolean_mask(x,not_xyz_mask,axis=1)],axis=1),tf.constant(idx_xyz_mask),axis=1)
    #simple_cost_recon = -1.0*tf.reduce_mean(mvn_r2.log_prob(new_x))
    #simple_cost_recon = -1.0*tf.reduce_mean(mvn_r2.log_prob(x))
    simple_cost_recon = -1.0*tf.reduce_mean(mvn_r2.log_prob(x))
    #simple_cost_recon = -1.0*tf.reduce_mean(tf.reduce_sum(gm_r2.log_prob(x),axis=1))

    selfent_q = -1.0*tf.reduce_mean(mvn_q.entropy())
    log_r1_q = gm_r1.log_prob(z_samp)   # evaluate the log prob of r1 at the q samples
    cost_KL = selfent_q - tf.reduce_mean(log_r1_q)
    return simple_cost_recon, cost_KL

    # TRUNCATED GAUSSIAN
    loc_gauss = tf.boolean_mask(mean_r2,gauss_mask,axis=1)
    scale_gauss = tf.boolean_mask(scale_r2,gauss_mask,axis=1)
    delta = tf.convert_to_tensor(10.0*(1.0-ramp), dtype=tf.float32)    # evolve boundaries 
    gauss = tfd.TruncatedNormal(loc_gauss,scale_gauss,-1.0*delta,1.0 + delta)   # shrink the truncation with the ramp
    log_prob_gauss = tf.reduce_sum(gauss.log_prob(tf.boolean_mask(x,gauss_mask,axis=1)),axis=1)

    # CONDITIONAL MASS
    mean_m1 = tf.boolean_mask(mean_r2,m1_mask,axis=1)
    mean_m2 = tf.boolean_mask(mean_r2,m2_mask,axis=1)
    scale_m1 = tf.boolean_mask(scale_r2,m1_mask,axis=1)
    scale_m2 = tf.boolean_mask(scale_r2,m2_mask,axis=1)
    delta = tf.convert_to_tensor(10.0*(1.0-ramp), dtype=tf.float32)     # evolve boundaries
    joint = tfd.JointDistributionSequential([    # shrink the truncation with the ramp
               tfd.TruncatedNormal(mean_m1,scale_m1,-1.0*delta,1.0 + delta,validate_args=True,allow_nan_stats=False), #reinterpreted_batch_ndims=None),  # m1
               lambda m1: tfd.TruncatedNormal(mean_m2,scale_m2,-1.0*delta,m1 + delta,validate_args=True,allow_nan_stats=False)],    # m2
        validate_args=True)
    log_prob_masses = joint.log_prob((tf.boolean_mask(x,m1_mask,axis=1),tf.boolean_mask(x,m2_mask,axis=1)))

    # VON-MISES
    delta = tf.convert_to_tensor(0.1 + 0.9*ramp, dtype=tf.float32)
    mean_vonmise = (tf.boolean_mask(mean_r2,vonmise_mask,axis=1) - 0.5)*delta + 0.5
    scale_vonmise = tf.boolean_mask(scale_r2,vonmise_mask,axis=1)*delta 
    scaled_x = (tf.boolean_mask(x,vonmise_mask,axis=1) - 0.5)*delta + 0.5    # shrink the parameter space into the centre of the VM
    con = tf.reshape(tf.square(tf.math.reciprocal(scale_vonmise)),[-1,vonmise_len])   # modelling wrapped scale output as log variance
    vonmise = tfp.distributions.VonMises(loc=2.0*np.pi*mean_vonmise, concentration=con)
    log_prob_vonmise = tf.reduce_sum(vonmise.log_prob(2.0*np.pi*tf.reshape(scaled_x,[-1,vonmise_len])),axis=1)

    # SKY
    #mean_sky = tf.boolean_mask(mean_r2,sky_mask,axis=1) - 0.5   # centre the domain around zero
    #scale_sky = tf.boolean_mask(scale_r2,ra_mask[:-1],axis=1)   # just use the ra mask
    #con = tf.reshape(tf.square(tf.math.reciprocal(scale_sky)),[-1])   # modelling wrapped scale output as log variance - only 1 concentration parameter for all sky
    #von_mises_fisher = tfp.distributions.VonMisesFisher(
    #                      mean_direction=tf.math.l2_normalize(tf.reshape(mean_sky,[-1,3]),axis=1),
    #                      concentration=con)   # define p_vm(2*pi*mu,con=1/sig^2)
    #ra = 2*np.pi*tf.reshape(tf.boolean_mask(x,ra_mask[:-1],axis=1),[-1,1])       # convert the scaled 0->1 true RA value back to radians
    #dec = np.pi*(tf.reshape(tf.boolean_mask(x,dec_mask[:-1],axis=1),[-1,1]) - 0.5) # convert the scaled 0>1 true dec value back to radians
    #xyz_unit = tf.reshape(tf.concat([tf.cos(ra)*tf.cos(dec),tf.sin(ra)*tf.cos(dec),tf.sin(dec)],axis=1),[-1,3])   # construct the true parameter unit vector
    #log_prob_sky = von_mises_fisher.log_prob(tf.math.l2_normalize(xyz_unit,axis=1))   # normalise it for safety (should already be normalised) and compute the logprob   
    #gauss_sky = tfp.distributions.MultivariateNormalDiag(
    #                      loc=tf.boolean_mask(mean_r2,xyz_mask,axis=1),
    #                      scale_diag=tf.boolean_mask(scale_r2,xyz_mask,axis=1))
    #log_prob_sky = gauss_sky.log_prob(xyz_actual)

    cost_recon = -1.0*tf.reduce_mean(log_prob_gauss + log_prob_masses + log_prob_vonmise) # + log_prob_sky)
    #cost_recon = simple_cost_recon*(1.0-ramp) + adv_cost_recon*ramp     # transition from simple to advanced
    #log_q_q = mvn_q.log_prob(z_samp)
    selfent_q = -1.0*tf.reduce_mean(mvn_q.entropy())
    log_r1_q = gm_r1.log_prob(z_samp)   # evaluate the log prob of r1 at the q samples
    #cost_KL = tf.reduce_mean(log_q_q - log_r1_q)
    cost_KL = selfent_q - tf.reduce_mean(log_r1_q)
    return cost_recon, cost_KL

def gen_samples(model, y, ramp=1.0, nsamples=1000, max_samples=1000):

    y = y/params['y_normscale']
    y = tf.tile(y,(max_samples,1,1))
    samp_iterations = int(nsamples/max_samples)
    for i in range(samp_iterations):
        mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
        scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
        gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1,
            scale_diag=scale_r1))
        z_samp = gm_r1.sample()
        mean_r2, logvar_r2 = model.decode_r2(z=z_samp,y=y)
        scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))
        #gm_r2 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r2),
        #    components_distribution=tfd.MultivariateNormalDiag(
        #    loc=mean_r2,
        #    scale_diag=scale_r2))
        mvn_r2 = tfp.distributions.MultivariateNormalDiag(
            loc=mean_r2,
            scale_diag=scale_r2)
        #gm_r2 = tfp.distributions.Beta(
        #    concentration1=mean_r2+EPS,
        #    concentration0=logvar_r2+EPS,
        #    validate_args=True, allow_nan_stats=False)
      
        if i==0:
            x_sample = mvn_r2.sample()
        else:
            x_sample = tf.concat([x_sample,mvn_r2.sample()],axis=0)
    return x_sample

#def gen_samples(model, y, ramp=1.0, nsamples=1000):
#    
#    y = y/params['y_normscale']
#    y = tf.tile(y,(nsamples,1,1))
#    mean_r1, logvar_r1, logweight_r1 = model.encode_r1(y=y)
#    scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
#    gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
#            components_distribution=tfd.MultivariateNormalDiag(
#            loc=mean_r1,
#            scale_diag=scale_r1))
#    z_samp = gm_r1.sample()
#    mean_r2, logvar_r2, logweight_r2 = model.decode_r2(z=z_samp,y=y)
#    scale_r2 = EPS + tf.sqrt(tf.exp(logvar_r2))
#
#    #mvn_r2 = tfp.distributions.MultivariateNormalDiag(
#    #                      loc=mean_r2,
#    #                      scale_diag=scale_r2)
#    #x_sample = mvn_r2.sample()    
#    gm_r2 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r2),
#            components_distribution=tfd.MultivariateNormalDiag(
#            loc=mean_r2,
#            scale_diag=scale_r2))
#    x_sample = gm_r2.sample()
#
#    #delta = tf.convert_to_tensor(10.0*(1.0-boundary), dtype=tf.float32)
#    #mvn_r2 = tfd.TruncatedNormal(mean_r2,scale_r2,-1.0*delta,1.0 + delta)   # shrink the truncation with the ramp
#    #x_sample = mvn_r2.sample()
#
#    # LOCALISATION
#    #xyz = tf.reshape(tf.boolean_mask(x_sample,xyz_mask,axis=1),[-1,3])          # sample the distribution
#    #samp_dist = tf.reshape(tf.math.reduce_euclidean_norm(xyz,axis=1),[-1,1])
#    #normed_xyz = tf.math.l2_normalize(xyz,axis=1)
#    #samp_ra = tf.math.floormod(tf.atan2(tf.slice(normed_xyz,[0,1],[-1,1]),tf.slice(normed_xyz,[0,0],[-1,1])),2.0*np.pi)/(2.0*np.pi)   # convert to the rescaled 0->1 RA from the unit vector
#    #samp_dec = (tf.asin(tf.slice(normed_xyz,[0,2],[-1,1])) + 0.5*np.pi)/np.pi                       # convert to the rescaled 0->1 dec from the unit vector
#    #x_samp_sky = tf.reshape(tf.concat([samp_dist,samp_ra,samp_dec],axis=1),[-1,3])             # group the sky samples
#    #new_x = tf.gather(tf.concat([x_samp_sky, tf.boolean_mask(x_sample,not_xyz_mask,axis=1)],axis=1),tf.constant(idx_xyz_mask),axis=1)
#    #return new_x
#    return x_sample
#    
#    # GAUSSIAN PARAMS
#    loc_gauss = tf.boolean_mask(mean_r2,gauss_mask,axis=1)
#    scale_gauss = tf.boolean_mask(scale_r2,gauss_mask,axis=1)
#    delta = tf.convert_to_tensor(10.0*(1.0-ramp), dtype=tf.float32)
#    gauss = tfd.TruncatedNormal(loc_gauss,scale_gauss,-1.0*delta,1.0+delta)   # shrink the truncation with the ramp
#    x_samp_gauss = gauss.sample()
#
#    # CONDITIONAL MASS
#    mean_m1 = tf.boolean_mask(mean_r2,m1_mask,axis=1)
#    mean_m2 = tf.boolean_mask(mean_r2,m2_mask,axis=1)
#    scale_m1 = tf.boolean_mask(scale_r2,m1_mask,axis=1)
#    scale_m2 = tf.boolean_mask(scale_r2,m2_mask,axis=1)
#    delta = tf.convert_to_tensor(10.0*(1.0-ramp), dtype=tf.float32)
#    joint = tfd.JointDistributionSequential([    # shrink the truncation with the ramp
#               tfd.TruncatedNormal(mean_m1,scale_m1,-1.0*delta,1.0+delta,validate_args=True,allow_nan_stats=False), #reinterpreted_batch_ndims=None),  # m1
#               lambda m1: tfd.TruncatedNormal(mean_m2,scale_m2,-1.0*delta,m1+delta,validate_args=True,allow_nan_stats=False)],    # m2
#        validate_args=True)
#    x_samp_masses = tf.transpose(tf.reshape(joint.sample(),[2,-1]))
#
#    # VON-MISES
#    delta = tf.convert_to_tensor(0.1 + 0.9*ramp, dtype=tf.float32)
#    mean_vonmise = (tf.boolean_mask(mean_r2,vonmise_mask,axis=1) - 0.5)*delta + 0.5
#    scale_vonmise = tf.boolean_mask(scale_r2,vonmise_mask,axis=1)*delta
#    con = tf.reshape(tf.square(tf.math.reciprocal(scale_vonmise)),[-1,vonmise_len])   # modelling wrapped scale output as log variance
#    vonmise = tfp.distributions.VonMises(loc=2.0*np.pi*mean_vonmise, concentration=con)
#    x_samp_vonmise = tf.math.floormod(vonmise.sample(),(2.0*np.pi))/(2.0*np.pi)
#    
#    # SKY
#    #gauss_sky = tfp.distributions.MultivariateNormalDiag(
#    #                      loc=tf.boolean_mask(mean_r2,xyz_mask,axis=1),
#    #                      scale_diag=tf.boolean_mask(scale_r2,xyz_mask,axis=1))
#    #xyz = gauss_sky.sample()          # sample the distribution
#    #samp_dist = tf.reshape(tf.math.reduce_euclidean_norm(xyz,axis=1),[-1,1])
#    #normed_xyz = tf.math.l2_normalize(xyz,axis=1)
#    #samp_ra = tf.math.floormod(tf.atan2(tf.slice(normed_xyz,[0,1],[-1,1]),tf.slice(normed_xyz,[0,0],[-1,1])),2.0*np.pi)/(2.0*np.pi)   # convert to the rescaled 0->1 RA from the unit vector
#    #samp_dec = (tf.asin(tf.slice(normed_xyz,[0,2],[-1,1])) + 0.5*np.pi)/np.pi                       # convert to the rescaled 0->1 dec from the unit vector
#    #x_samp_sky = tf.reshape(tf.concat([samp_dist,samp_ra,samp_dec],axis=1),[-1,3])             # group the sky samples
#    #mean_sky = tf.boolean_mask(mean_r2,sky_mask,axis=1) - 0.5    # centre the domain around zero
#    #scale_sky = tf.boolean_mask(scale_r2,ra_mask[:-1],axis=1)   # just use the ra mask
#    #con = tf.reshape(tf.square(tf.math.reciprocal(scale_sky)),[-1])   # modelling wrapped scale output as log variance - only 1 concentration parameter for all sky
#    #von_mises_fisher = tfp.distributions.VonMisesFisher(
#    #                      mean_direction=tf.math.l2_normalize(tf.reshape(mean_sky,[-1,3]),axis=1),
#    #                      concentration=con)   # define p_vm(2*pi*mu,con=1/sig^2)
#    #xyz = tf.reshape(von_mises_fisher.sample(),[-1,3])          # sample the distribution
#    #samp_ra = tf.math.floormod(tf.atan2(tf.slice(xyz,[0,1],[-1,1]),tf.slice(xyz,[0,0],[-1,1])),2.0*np.pi)/(2.0*np.pi)   # convert to the rescaled 0->1 RA from the unit vector
#    #samp_dec = (tf.asin(tf.slice(xyz,[0,2],[-1,1])) + 0.5*np.pi)/np.pi                       # convert to the rescaled 0->1 dec from the unit vector
#    #x_samp_sky = tf.reshape(tf.concat([samp_ra,samp_dec],axis=1),[-1,2])             # group the sky samples
#
#    return tf.gather(tf.concat([x_samp_masses, x_samp_gauss, x_samp_vonmise],axis=1),tf.constant(idx_mask),axis=1)

def ramp_func(epoch,start,ramp_length, n_cycles):
    i = (epoch-start)/(2.0*ramp_length)
    if i<0:
        return 0.0
    if i>=n_cycles:
        return 1.0
    return min(1.0,2.0*np.remainder(i,1.0))

def gen_z_samples(model, x, y, nsamples=1000,max_samples=1000):

    y = y/params['y_normscale']
    y = tf.tile(y,(max_samples,1,1))
    x = tf.tile(x,(max_samples,1))
    samp_iterations = int(nsamples/max_samples)
    for i in range(samp_iterations):

        mean_r1_temp, logvar_r1, logweight_r1 = model.encode_r1(y=y)
        scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
        gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=mean_r1_temp,
            scale_diag=scale_r1))
        mean_q_temp, logvar_q = model.encode_q(x=x,y=y)
        scale_q = EPS + tf.sqrt(tf.exp(logvar_q))
        mvn_q = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_q_temp,
                          scale_diag=scale_q)
        if i==0:
            z_samp_q = mvn_q.sample()
            z_samp_r1 = gm_r1.sample()
            mean_r1 = mean_r1_temp
            mean_q = mean_q_temp
        else:
            z_samp_q = tf.concat([z_samp_q,mvn_q.sample()],axis=0)
            z_samp_r1 = tf.concat([z_samp_r1,gm_r1.sample()],axis=0)
            mean_r1 = tf.concat([mean_r1,mean_r1_temp],axis=0)
            mean_q = tf.concat([mean_q,mean_q_temp],axis=0)
    
    return mean_r1, z_samp_r1, mean_q, z_samp_q

@tf.function
def train_step(model, x, y, optimizer, ramp=1.0):
    """Executes one training step and returns the loss.
    This function computes the loss and gradients, and uses the latter to
    update the model's parameters.
    """
    with tf.GradientTape() as tape:
        r_loss, kl_loss = compute_loss(model, x, y, ramp=ramp)
        loss = r_loss + ramp*kl_loss
    gradients = tape.gradient(loss, model.trainable_variables)
    optimizer.apply_gradients(zip(gradients, model.trainable_variables))
    train_loss_metric(loss)
    return r_loss, kl_loss

epochs = 25000
train_loss_metric = tf.keras.metrics.Mean('train_loss', dtype=tf.float32)
train_log_dir = '/data/www.astro/chrism/vitamin_c/' + run + '/logs'
#os.mkdir(train_log_dir)
train_summary_writer = tf.summary.create_file_writer(train_log_dir)
model = CVAE(x_data_train.shape[1], params['ndata'], y_data_train.shape[2], params['z_dimension'], params['n_modes'])

# start the training loop
train_loss = np.zeros((epochs,3))
val_loss = np.zeros((epochs,3))
ramp_start = 200
#ramp_stop = 200
ramp_length = 200
#ramp_grad = 1.0/(ramp_stop - ramp_start)
ramp_cycles = 1
#boundary_start = 10
#boundary_stop = 20
#boundary_grad = 1.0/(boundary_stop - boundary_start)
KL_samples = []
for epoch in range(1, epochs + 1):

    train_loss_kl_q = 0.0
    train_loss_kl_r1 = 0.0
    start_time_train = time.time()
    ramp = tf.convert_to_tensor(ramp_func(epoch,ramp_start,ramp_length,ramp_cycles), dtype=tf.float32)
    #boundary = tf.convert_to_tensor(np.min([(epoch-boundary_start)*boundary_grad,1.0]).astype(np.single) if epoch>boundary_start else 0.0, dtype=tf.float32)
    for step, (x_batch_train, y_batch_train) in train_dataset.enumerate():
        temp_train_r_loss, temp_train_kl_loss = train_step(model, x_batch_train, y_batch_train, optimizer, ramp=ramp)
        train_loss[epoch-1,0] += temp_train_r_loss
        train_loss[epoch-1,1] += temp_train_kl_loss
    train_loss[epoch-1,2] = train_loss[epoch-1,0] + ramp*train_loss[epoch-1,1]
    train_loss[epoch-1,:] /= float(step+1)
    end_time_train = time.time()
    with train_summary_writer.as_default():
        tf.summary.scalar('loss', train_loss_metric.result(), step=epoch)
    train_loss_metric.reset_states()

    start_time_val = time.time()
    for step, (x_batch_val, y_batch_val) in val_dataset.enumerate():
        temp_val_r_loss, temp_val_kl_loss = compute_loss(model, x_batch_val, y_batch_val, ramp=ramp)
        val_loss[epoch-1,0] += temp_val_r_loss
        val_loss[epoch-1,1] += temp_val_kl_loss
    val_loss[epoch-1,2] = val_loss[epoch-1,0] + ramp*val_loss[epoch-1,1]
    val_loss[epoch-1,:] /= float(step+1)
    end_time_val = time.time()

    print('Epoch: {}, Run {}, Training RECON: {}, KL: {}, TOTAL: {}, time elapsed: {}'
        .format(epoch, run, train_loss[epoch-1,0], train_loss[epoch-1,1], train_loss[epoch-1,2], end_time_train - start_time_train))
    print('Epoch: {}, Run {}, Validation RECON: {}, KL: {}, TOTAL: {}, time elapsed {}'
        .format(epoch, run, val_loss[epoch-1,0], val_loss[epoch-1,1], val_loss[epoch-1,2], end_time_val - start_time_val))

    # update loss plot
    plot_losses(train_loss, val_loss, epoch, run=run)

    # generate and plot posterior samples for the latent space and the parameter space 
    if epoch % plot_cadence == 0:
        for step, (x_batch_test, y_batch_test, snrs_batch_test) in test_dataset.enumerate():             
            mu_r1, z_r1, mu_q, z_q = gen_z_samples(model, x_batch_test, y_batch_test, nsamples=8000)
            plot_latent(mu_r1,z_r1,mu_q,z_q,epoch,step,run=run)
            start_time_test = time.time()
            samples = gen_samples(model, y_batch_test, ramp=ramp, nsamples=8000)
            end_time_test = time.time()
            if np.any(np.isnan(samples)):
                print('Epoch: {}, found nans in samples. Not making plots'.format(epoch))
                for k,s in enumerate(samples):
                    if np.any(np.isnan(s)):
                        print(k,s)
                KL_est = [-1,-1,-1]
            else:
                print('Epoch: {}, run {} Testing time elapsed for 8000 samples: {}'.format(epoch,run,end_time_test - start_time_test))
                KL_est = plot_posterior(samples,x_batch_test[0,:],snrs_batch_test[0,:],epoch,step,all_other_samples=bilby_samples[:,step,:],run=run)
                _ = plot_posterior(samples,x_batch_test[0,:],snrs_batch_test[0,:],epoch,step,run=run)
            KL_samples.append(KL_est)

        # make p-p plot
        results = []
        for step, (x_batch_pp, y_batch_pp, snrs_batch_pp) in pp_dataset.enumerate():   #for each datafile, cast the true values and generated samples into an dictionary
            posterior = dict()
            injections = dict()
            priors = dict()
            samples = gen_samples(model,y_batch_pp,ramp=ramp,nsamples=1000)
            x_batch_pp = x_batch_pp.numpy().astype('float64')
            for i,x in enumerate(x_batch_pp[0]):
                posterior[str(i)] = samples[:,i]
                injections[str(i)] = x
                priors[str(i)] = Uniform(-100,100)
            posterior = pd.DataFrame(dict(posterior))
            result = bilby.result.Result(
                label="test",
                injection_parameters=injections,
                posterior=posterior,
                search_parameter_keys=injections.keys(),
                priors=priors)
            results.append(result)
        bilby.result.make_pp_plot(results, filename='/data/www.astro/chrism/vitamin_c/%s/pp_plot_epoch_%d.png' % (run,epoch),
                              confidence_interval=[0.68,0.95,0.997])

        # plot KL evolution
        #plot_KL(np.reshape(np.array(KL_samples),[-1,4,3]),plot_cadence,run=run)

    # load more noisefree training data back in
    if epoch % 25 == 0:
        x_data_train, y_data_train, _, snrs_train = load_data(params,bounds,fixed_vals,params['train_set_dir'],params['inf_pars'],silent=True)

        # augment the data for this chunk - randomize phase and arrival time
        old_phase = bounds['phase_min'] + tf.boolean_mask(x_data_train,phase_mask,axis=1)*(bounds['phase_max'] - bounds['phase_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_phase = bounds['phase_min'] + new_x*(bounds['phase_max'] - bounds['phase_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_phase_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
        old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x_data_train,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
        new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
        x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x_data_train,not_geocent_mask,axis=1),[-1,tf.shape(x_data_train)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
        phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
        phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y_data_train)[1]/2 + 1))
        fvec = tf.range(params['ndata']/2 + 1)/params['duration']
        time_correction = -2.0*np.pi*fvec*(new_geocent-old_geocent)
        time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
        time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
        y_data_train_fft = tf.signal.rfft(tf.transpose(y_data_train,[0,2,1]))*phase_correction*time_correction
        y_data_train = tf.transpose(tf.signal.irfft(y_data_train_fft),[0,2,1])

        train_dataset = (tf.data.Dataset.from_tensor_slices((x_data_train,y_data_train))
                 .shuffle(train_size).batch(batch_size))





















